#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A script to install and start an SSH daemon in a Docker image, enabling the
user to log on to it.
"""

import argparse
import os
import pwd
import socket
import subprocess
import sys
import time
import yaml
# pylint:disable=no-name-in-module,import-error
from distutils.spawn import find_executable
# pylint:enable=no-name-in-module,import-error

from xdg.BaseDirectory import xdg_config_home


from forklift.arguments import argument_factory, convert_to_args, project_args
from forklift.base import DEVNULL, ImproperlyConfigured
import forklift.drivers
import forklift.services


def create_parser(services, drivers, command_required=True):
    """
    Collect all options from services and drivers in an argparse format.
    """

    parser = argparse.ArgumentParser(
        usage="%(prog)s [options]",
    )
    add_argument = parser.add_argument

    add_argument('--application_id',
                 help="Application name to derive resource names from")
    add_argument('--driver', default=None, choices=drivers.keys(),
                 help="Driver to execute the application with")
    add_argument('--services', default=[], nargs='*', choices=services.keys(),
                 help="Services to provide to the application")
    add_argument('--environment', default=[], nargs='*',
                 type=lambda pair: pair.split('=', 1),
                 help="Additional environment variables to pass")

    for name, service in services.items():
        service_options = parser.add_argument_group(name)
        service.add_arguments(
            argument_factory(service_options.add_argument, name))

    add_argument('command', nargs='+' if command_required else '*',
                 help="Command to run")

    # Drivers inherit all the common options from their base class, so
    # allow conflicts for this group of options
    driver_options = parser.add_argument_group('Driver options')
    driver_options.conflict_handler = 'resolve'
    for name, driver in drivers.items():
        driver.add_arguments(driver_options.add_argument)

    # Dummy option to separate command line arguments from the ones
    # generated from configuration files
    add_argument('--zzzz', action='store_const', const=None,
                 help=argparse.SUPPRESS)

    return parser


class Forklift(object):
    """
    The main class.
    """

    services = forklift.services.register
    drivers = forklift.drivers.register

    CONFIG_DIR = os.path.join(xdg_config_home, 'forklift')

    def __init__(self, argv):
        """
        Parse the command line and set up the class.
        """

        # Parse the configuration from:
        # - implicit defaults
        # - project configuration file
        # - user configuration file
        # - user per-project configuration file
        # - command line

        options = self.implicit_configuration()

        # Get application_id
        initial_parser = create_parser({}, {}, command_required=False)
        conf, _ = initial_parser.parse_known_args(options)

        for conffile in self.configuration_files(conf):
            options.extend(self.file_configuration(conffile))

        options.append('--zzzz')
        options.extend(argv[1:])

        parser = create_parser(self.services, self.drivers)

        conf = parser.parse_args(options)

        # Once the driver and services are known, parse the arguments again
        # with only the needed options

        driver = self.get_driver(conf)
        # enabled_services = {
        #     name: service
        #     for name, service in self.services.items()
        #     if name in conf.services
        # }

        # FIXME: creating a parser with only the enabled_services (see above)
        # causes problems because we then cannot parse the arguments for
        # disabled services. Because services are separately namespaced
        # including arguments for non-enabled services is sufficient for now
        parser = create_parser(self.services,  # FIXME: enabled_services
                               {driver: self.drivers[driver]})

        self.conf = parser.parse_args(options)

    def implicit_configuration(self):
        """
        Implicit configuration based on the current directory.
        """

        application_id = os.path.basename(os.path.abspath(os.curdir))
        return [
            '--application_id', application_id,
        ]

    def configuration_files(self, conf):
        """
        A list of configuration files to look for settings in.
        """

        application_id = conf.application_id
        return (
            'forklift.yaml',
            os.path.join(self.CONFIG_DIR, '_default.yaml'),
            os.path.join(self.CONFIG_DIR, '{0}.yaml'.format(application_id)),
        )

    def file_configuration(self, name):
        """
        Parse settings from a configuration file.
        """
        try:
            with open(name) as conffile:
                return convert_to_args(yaml.load(conffile))
        except IOError:
            return []

    @staticmethod
    def _readme_stream():
        """
        Get the README file as a stream.
        """

        from pkg_resources import resource_stream
        return resource_stream(__name__, 'README.md')

    def help(self):
        """
        Render the help file.
        """

        readme = self._readme_stream()

        # Try to format the README nicely if Pandoc is installed
        pagers = [
            'pandoc -s -f markdown -t man | man -l -',
            os.environ.get('PAGER', ''),
            'less',
            'more',
        ]

        pager = None

        for pager in pagers:
            if find_executable(pager.split(' ')[0]):
                break

        process = subprocess.Popen(pager, shell=True, stdin=subprocess.PIPE)
        process.communicate(input=readme.read())
        readme.close()
        process.wait()

    def get_driver(self, conf):
        """
        Find out what driver to use given the configuration.

        If no driver is explicitly specified, choose one which states
        the command is its valid target or fall back to Docker driver.
        """

        if conf.driver:
            return conf.driver

        target = conf.command[0]
        for driver_name, driver_class in self.drivers.items():
            if driver_class.valid_target(target):
                return driver_name

        return 'docker'

    def main(self):
        """
        Run the specified application command.
        """

        if self.conf.command == ['help']:
            self.help()
            return 0

        driver_name = self.get_driver(self.conf)
        driver_class = self.drivers[driver_name]

        (target, *command) = self.conf.command

        try:
            services = [
                self.services[service].provide(
                    self.conf.application_id,
                    project_args(self.conf, service)
                )
                for service in self.conf.services
            ]

            environment = dict(self.conf.environment)

            driver = driver_class(
                target=target,
                services=services,
                environment=environment,
                conf=self.conf,
            )

        except ImproperlyConfigured as ex:
            print(ex)
            return 1

        return driver.run(*command)


def main():
    """
    Main entry point.
    """

    return Forklift(sys.argv).main()


if __name__ == '__main__':
    main()



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test SSH daemon setup.
"""

import os
import subprocess

from tests.base import (
    docker,
    DOCKER_BASE_IMAGE,
    merge_dicts,
    TestCase,
    TestForklift,
)

from forklift.base import DEVNULL
from forklift.drivers import Docker


class SaveSSHDetailsDocker(Docker):
    """
    Save SSH command the container ran.
    """

    log = []

    def ssh_command(self, container, identity=None):
        """
        Save the SSH command for later inspection.
        """

        command, available = super().ssh_command(container, identity=identity)
        self.log.append((command, available, container))
        return command, available

    @classmethod
    def last_details(cls):
        """
        Return the next (FIFO) SSH command.
        """

        return cls.log.pop(0)


class SSHTestForklift(TestForklift):
    """
    Test Forklift saving SSH commands.
    """

    drivers = merge_dicts({
        'save_ssh_command_docker': SaveSSHDetailsDocker,
    }, TestForklift.drivers)


@docker
class SSHTestCase(TestCase):
    """
    Test setting up an SSH daemon via Docker.
    """

    private_key = 'tests/test_id_rsa'

    forklift_class = SSHTestForklift

    def test_sshd(self):
        """
        Test setting up an SSH daemon.
        """

        container = None
        try:
            os.chmod(self.private_key, 0o600)

            self.assertEqual(0, self.run_forklift(
                '--driver', 'save_ssh_command_docker',
                DOCKER_BASE_IMAGE, 'sshd',
                '--identity', self.private_key,
            ))

            command, available, container = SaveSSHDetailsDocker.last_details()

            self.assertTrue(available)
            self.assertEqual(0, subprocess.call(
                command + ' /bin/true',
                shell=True
            ))
        finally:
            # Kill and remove the started container
            if container is not None:
                for action in ('stop', 'rm'):
                    subprocess.check_call(
                        ('docker', action, container),
                        stdout=DEVNULL,
                    )



"""
Argument parsing utilities.
"""

from argparse import Namespace


def argument_factory(add_argument, name):
    """
    A factory to prepend all argument names with a prefix.
    """

    def wrapped(*args, **kwargs):
        """
        Prepend all argument names with a prefix before adding the argument.
        """

        assert all(option[:2] == '--' for option in args)
        option_names = [
            '--{0}.{1}'.format(name, option[2:])
            for option in args
        ]
        return add_argument(*option_names, **kwargs)

    return wrapped


def convert_to_args(conf, prefix=None):
    """
    Convert hierarchical configuration dictionary to argparse arguments.

    'environment' at root level is a special case: if it is a hash,
    it is converted into an array of VAR=VALUE pairs.
    """

    args = []
    conf = conf or {}
    prefix = prefix or ()

    if not prefix and 'environment' in conf:
        environment = conf['environment']
        if isinstance(environment, dict):
            conf['environment'] = [
                '{0}={1}'.format(k, v)
                for k, v in environment.items()
            ]

    for key, value in conf.items():
        arg_prefix = prefix + (key,)
        if isinstance(value, dict):
            args.extend(convert_to_args(value, arg_prefix))
        else:
            if not isinstance(value, (list, tuple)):
                value = (value,)

            if len(value) > 0:
                args.append('--' + '.'.join(arg_prefix))
                for val in value:
                    args.append(str(val))

    return args


def project_args(args, prefix):
    """
    Return only keys in the object having the specified prefix, stripping
    the prefix.
    """

    pairs = vars(args).items()
    strip_len = len(prefix) + 1

    return Namespace(**dict(
        (key[strip_len:], value)
        for key, value in pairs
        if key.startswith(prefix + '.')
    ))



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Setup script.
"""

from setuptools import setup, find_packages
from sys import version_info

assert version_info >= (3,), "Python 3 is required."

setup(
    name='docker-forklift',
    version='0.2.16',
    description='Utility for running a container',
    author='Infoxchange Australia development team',
    author_email='devs@infoxchange.net.au',
    url='https://github.com/infoxchange/docker-forklift',
    license='Apache 2.0',
    long_description=open('README.md').read(),

    packages=find_packages(exclude=['tests']),
    package_data={
        'forklift': ['README.md'],
    },
    entry_points={
        'console_scripts': [
            'forklift = forklift:main',
        ],
    },

    install_requires=[
        'pyxdg',
        'pyyaml',
    ],

    test_suite='tests',
    tests_require=[
        'pep8',
        'pylint',
        'pylint-mccabe',
    ],
)



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Services that can be provided to running applications.
"""

import json
import os
import socket
import subprocess
import urllib.request

from forklift.base import free_port, ImproperlyConfigured
from forklift.registry import Registry


def port_open(host, port):
    """
    Check whether the specified TCP port is open.
    """

    with socket.socket() as sock:
        try:
            sock.connect((host, int(port)))
            return True
        except socket.error:
            return False


def split_host_port(host_port, default_port):
    """
    Split host:port into host and port, using the default port in case
    it's not given.
    """

    host_port = host_port.split(':')
    if len(host_port) == 2:
        host, port = host_port
        return host, port
    else:
        return host_port[0], default_port


register = Registry()  # pylint:disable=invalid-name


class Service(object):
    """
    Base class for services required by the application.
    """

    # A list of class methods to try to find an available service provider.
    providers = ()

    # A list of attributes which can be overridden from a configuration file
    # or the command line.
    allow_override = ()

    # A list of attributes which can be overridden as a list of arguments
    # (i.e. hosts, urls)
    allow_override_list = ()

    @classmethod
    def add_arguments(cls, add_argument):
        """
        Add service configuration arguments to the parser.
        """

        # TODO: refactor for types other than string (port numbers) and
        # list (Elasticsearch host).

        for param in cls.allow_override:
            add_argument('--{0}'.format(param))

        for param in cls.allow_override_list:
            add_argument('--{0}'.format(param), nargs='+')

    @classmethod
    def provide(cls, application_id, overrides=None):
        """
        Choose the first available service from the list of providers.
        """

        overrides = overrides or {}
        allowed_overrides = cls.allow_override + cls.allow_override_list

        for provider in cls.providers:
            service = getattr(cls, provider)(application_id)

            for key, value in vars(overrides).items():
                if value is not None:
                    if key in allowed_overrides:
                        setattr(service, key, value)
                    else:
                        raise ImproperlyConfigured(
                            "Invalid parameter {0} for service {1}.".format(
                                key, cls.__name__))

            if service.available():
                return service

        raise ImproperlyConfigured(
            "No available providers for service {0}.".format(cls.__name__))

    def available(self):
        """
        Check whether the service is available. Override to implement
        availability checks to warn the user instead of let the application
        fail.
        """

        return True

    def environment(self):
        """
        The environment, as a dictionary, to let the application know
        the service configuration.
        """

        raise NotImplementedError("Please override environment().")


@register('postgres')
class PostgreSQLService(Service):
    """
    PostgreSQL service provided by the host machine.
    """

    DATABASE_NAME = 'DEFAULT'

    URL_SCHEME = 'postgres'

    CHECK_COMMAND = 'select version()'

    allow_override = ('name', 'host', 'port', 'user', 'password')

    # pylint:disable=too-many-arguments
    def __init__(self,
                 name,
                 host='localhost',
                 port=5432,
                 user=None,
                 password=None):
        self.host = host
        self.port = port
        self.name = name
        self.user = user
        self.password = password

    def environment(self):
        """
        The environment needed for the application to connect to PostgreSQL.
        """

        env_name = 'DB_{0}_URL'.format(self.DATABASE_NAME)
        details = {k: v or '' for k, v in self.__dict__.items()}
        details['scheme'] = self.URL_SCHEME
        url = '{scheme}://{user}:{password}@{host}:{port}/{name}'.format(
            **details)
        return {env_name: url}

    def available(self):
        """
        Check whether PostgreSQL is installed on the host and accessible.
        """

        try:
            subprocess.check_output(['psql', '--version'])

            if self.password:
                os.environ['PGPASSWORD'] = self.password
            subprocess.check_output([
                'psql',
                '-h', self.host,
                '-p', str(self.port),
                '-U', self.user,
                '-w',
                self.name,
                '-c', self.CHECK_COMMAND,
            ])

            return True
        except subprocess.CalledProcessError:
            return False

    @classmethod
    def localhost(cls, application_id):
        """
        The PostgreSQL environment on the local machine.
        """
        return cls(
            host='localhost',
            name=application_id,
            user=application_id,
        )

    providers = ('localhost',)


@register('postgis')
class PostGISService(PostgreSQLService):
    """
    PostgreSQL database with PostGIS support.
    """

    URL_SCHEME = 'postgis'

    CHECK_COMMAND = 'select PostGIS_full_version()'


def pipe_split(value):
    """
    Split a pipe-separated string if it's the only value in an array.
    """

    if len(value) == 1 and '|' in value[0]:
        return value[0].split('|')
    else:
        return value


@register('memcache')
class MemcacheService(Service):
    """
    Memcache service for the application.
    """

    allow_override = ('key_prefix', 'host')
    allow_override_list = ('hosts',)
    providers = ('localhost',)

    DEFAULT_PORT = 11211

    def __init__(self,
                 key_prefix='',
                 hosts=None):

        self.key_prefix = key_prefix
        self.hosts = hosts or []

    def environment(self):
        """
        The environment to access Memcache
        """

        return {
            'MEMCACHE_HOSTS': '|'.join(self.hosts),
            'MEMCACHE_PREFIX': self.key_prefix,
        }

    def available(self):
        """
        Check whether memcache is available

        Do this by connecting to the socket. At least one host must be up
        """

        if not self.hosts:
            return False

        for host in self.hosts:
            if port_open(*split_host_port(host, self.DEFAULT_PORT)):
                return True

        return False

    @property
    def host(self):
        """
        The (pipe separated) hosts for the Memcache service.
        """

        return '|'.join(
            split_host_port(host, self.DEFAULT_PORT)[0]
            for host in self.hosts
        )

    @host.setter
    def host(self, host):
        """
        Set the host to access Memcache at.
        """

        self.hosts = [
            h if ':' in h else '{0}:{1}'.format(h, self.DEFAULT_PORT)
            for h in host.split('|')
        ]

    @classmethod
    def localhost(cls, application_id):
        """
        The default memcached provider
        """

        return cls(key_prefix=application_id,
                   hosts=['localhost:{0}'.format(cls.DEFAULT_PORT)])


@register('elasticsearch')
class ElasticsearchService(Service):
    """
    Elasticsearch service for the application.
    """

    allow_override = ('index_name', 'host')
    allow_override_list = ('urls',)

    def __init__(self, index_name, urls):
        self.index_name = index_name
        self._url_array = []
        self.urls = urls

    def environment(self):
        """
        The environment to access Elasticsearch.
        """

        return {
            'ELASTICSEARCH_INDEX_NAME': self.index_name,
            'ELASTICSEARCH_URLS': self.url_string(),
        }

    def url_string(self):
        """
        All URLs joined as a string.
        """
        return '|'.join(url.geturl() for url in self.urls)

    @property
    def urls(self):
        """
        The (pipe separated) URLs to access Elasticsearch at.
        """

        return self._url_array

    @urls.setter
    def urls(self, urls):
        """
        Set the URLs to access Elasticsearch at.
        """

        self._url_array = [
            urllib.parse.urlparse(url) if isinstance(url, str) else url
            for url in pipe_split(urls)
        ]

    @property
    def host(self):
        """
        The (pipe separated) hosts for the Elasticsearch service.
        """

        return '|'.join(url.hostname for url in self._url_array)

    @host.setter
    def host(self, host):
        """
        Set the host to access Elasticsearch at.
        """

        self.urls = [
            # pylint:disable=protected-access
            url._replace(
                netloc='{host}:{port}'.format(host=host, port=url.port))
            for url in self.urls
        ]

    def available(self):
        """
        Check whether Elasticsearch is available at a given URL.
        """

        if not self.urls:
            return False

        for url in self.urls:
            try:
                es_response = urllib.request.urlopen(url.geturl())
                es_status = json.loads(es_response.read().decode())
                if es_status['status'] != 200:
                    return False
            except (urllib.request.URLError, ValueError):
                return False

        return True

    @classmethod
    def localhost(cls, application_id):
        """
        The Elasticsearch environment on the local machine.
        """
        return cls(index_name=application_id,
                   urls=('http://localhost:9200',))

    providers = ('localhost',)


@register('proxy')
class ProxyService(Service):
    """
    Proxy service for the application.
    """

    allow_override = ('host', 'port')

    def __init__(self, host=None, port=None):
        self.host = host
        self.port = port

    def available(self):
        """
        Check whether the proxy is available.
        """

        if self.host:
            return port_open(self.host, self.port)
        else:
            return True

    def environment(self):
        """
        The environment to access the proxy.
        """

        if self.host:
            return {
                'HTTP_PROXY': 'http://{host}:{port}'.format(**self.__dict__),
            }
        else:
            return {}

    # pylint:disable=unused-argument
    @classmethod
    def manual(cls, application_id):
        """
        Manually-configured proxy. Will not be available unless parameters
        are overridden in configuration.
        """

        return cls()

    providers = ('manual',)


@register('email')
class EmailService(Service):
    """
    An MTA for the application.
    """

    allow_override = ('host', 'port')

    def __init__(self, host, port=25):
        self.host = host
        self.port = port

    def environment(self):
        """
        The environment to send email.
        """

        return {
            'EMAIL_HOST': self.host,
            'EMAIL_PORT': str(self.port),
        }

    def available(self):
        """
        Check whether the MTA is available.
        """

        return port_open(self.host, self.port)

    # pylint:disable=unused-argument
    @classmethod
    def localhost(cls, application_id):
        """
        The MTA on the local machine.
        """
        return cls(host='localhost')

    providers = ('localhost',)


@register('syslog')
class SyslogService(Service):
    """
    Logging facility for the application.
    """

    DEFAULT_PORT = 514

    allow_override = ('host', 'port', 'proto')

    def __init__(self, host=None, port=DEFAULT_PORT, proto='udp'):
        self.host = host
        self.port = port
        self.proto = proto

    def environment(self):
        """
        The environment to provide logging.
        """

        return {
            'SYSLOG_SERVER': self.host,
            'SYSLOG_PORT': str(self.port),
            'SYSLOG_PROTO': self.proto,
        }

    def available(self):
        """
        Check whether syslog is available.

        If the protocol is UDP, assume it is available if any of the other
        parameters are set.
        """

        if self.host is None:
            return False

        if self.proto == 'udp':
            return True
        else:
            return port_open(self.host, self.port)

    # pylint:disable=unused-argument
    @classmethod
    def manual(cls, application_id):
        """
        Manually-configured syslog. Will not be available unless parameters
        are overridden in configuration.
        """

        return cls()

    # pylint:disable=unused-argument
    @classmethod
    def stdout(cls, application_id):
        """
        Logger printing all the messages to the standard output of Forklift.
        """

        # Adapted from https://gist.github.com/marcelom/4218010
        import socketserver
        import threading

        class SyslogHandler(socketserver.BaseRequestHandler):
            """
            Handler outputting logging messages received to stdout.
            """
            def handle(self):
                data = self.request[0].strip().decode()
                print(data)

        class ThreadedUDPServer(socketserver.ThreadingMixIn,
                                socketserver.UDPServer):
            """
            Threaded UDP server.
            """
            pass

        port = free_port()
        server = ThreadedUDPServer(('0.0.0.0', port), SyslogHandler)
        server_thread = threading.Thread(target=server.serve_forever)
        server_thread.daemon = True
        server_thread.start()

        return cls('localhost', port)

    providers = ('manual', 'stdout')
